# -*- coding: utf-8 -*-
# Corpuscula project: Wrappers for known corpora
#
# Copyright (C) 2019-present by Sergei Ternovykh
# License: BSD, see LICENSE for details
"""
Tools for downloading and converting known corpora of Russian language.
Includes wrapper for corpora to simplify the further processing.
"""
from collections import OrderedDict
from html import unescape
import json
import os
from pathlib import Path
import re
import sys
from urllib.error import HTTPError

from corpuscula.conllu import Conllu
from corpuscula.utils import LOG_FILE, download_file, rmdir, \
                             read_bz2, read_rar, read_zip


ROOT_DIR = os.path.dirname(os.path.realpath(__file__))
_CFG_ROOT_DIR = 'ROOT_DIR'
def set_root_dir(root_dir):
    root_dir = root_dir.strip()
    cfg_path = os.path.join(str(Path.home()), '.rumor')
    cfg = ["# Config file for RuMor project. Don't change it manually."]
    isdone = False
    if os.path.isfile(cfg_path):
        with open(cfg_path, 'rt', encoding='utf-8') as f:
            for line in f:
                try:
                    var, val = line.split('\t', maxsplit=1)
                    if var == _CFG_ROOT_DIR:
                        if not isdone:
                            val = root_dir
                            isdone = True
                        else:
                            continue
                    cfg.append('\t'.join((var, val.strip())))
                except ValueError:
                    pass
    if not isdone:
        cfg.append('\t'.join((_CFG_ROOT_DIR, root_dir)))
    with open(cfg_path, 'wt', encoding='utf-8') as f:
        for line in cfg:
            print(line, file=f)

def get_root_dir():
    cfg_path = os.path.join(str(Path.home()), '.rumor')
    res = ROOT_DIR
    with open(cfg_path, 'rt', encoding='utf-8') as f:
        for line in f:
            try:
                var, val = line.split('\t', maxsplit=1)
                if var == _CFG_ROOT_DIR:
                    res = val.strip()
                    break
            except ValueError:
                pass
    return res

CORPUS_DNAME = 'corpus'
def download_corpus(name, url, dname=None, root_dir=None, fname=None,
                    **kwargs):
    """Download a corpus *name* from *url*. The corpus will be stored inside
    the project corpus storage in the directory *dname* with the file name
    *fname*.

    If *dname* is None, *name* param is used instead.

    If *fname is None, the name of the file will be copied from the *url*

    :param root_dir: path to the root storage. If None, default value will
                     be used
    :type root_dir: str

    Also, the function receives other parameters that fit for
    ``.utils.download_file()`` method"""
    dpath = os.path.join(root_dir if root_dir else get_root_dir(),
                         CORPUS_DNAME,
                         dname if dname is not None else name)
    
    return download_file(url, dpath=dpath, fname=fname,
                         log_msg='Downloading {}'.format(name), **kwargs)

def remove_corpus(dname, root_dir=None):
    """Remove a directory *dname* from the project corpus storage.

    WARNING: All files in the directory will be removed.

    :param root_dir: path to the root storage. If None, default value will
                     be used
    :type root_dir: str

    If *dname* is None, the corpus storage will be totally cleaned. All
    downloaded corpora will be removed"""
    dpath = os.path.join(root_dir if root_dir else get_root_dir(),
                         CORPUS_DNAME,
                         dname if dname is not None else '')
    try:
        rmdir(dpath)
    except FileNotFoundError:
        pass
    try:
        os.rmdir(os.path.join(root_dir if root_dir else get_root_dir(),
                              CORPUS_DNAME))
    except OSError:
        pass

def get_corpus_fpath(dname=None, root_dir=None, fname=None, url=None):
    """Return full path for file *fname* from project corpus storage directory
    *dname*. If *fname* is None, the name of the file will be copied from the
    *url*. No cheking of the file existance is made.

    :param root_dir: path to the root storage. If None, default value will
                     be used
    :type root_dir: str
    """
    assert fname is not None or url is not None, \
        '*fname* or *url* must be not None'
    dpath = os.path.join(root_dir if root_dir else get_root_dir(),
                         CORPUS_DNAME,
                         dname if dname is not None else '')
    fpath = os.path.join(dpath, fname if fname else
                                re.sub('^.+/', '', url))
    return fpath

def _get_ud_train_name(file_list):
    return next((x for x in file_list if x.endswith('train.conllu')), None)

def _get_ud_dev_name(file_list):
    return next((x for x in file_list if x.endswith('dev.conllu')), None)

def _get_ud_test_name(file_list):
    return next((x for x in file_list if x.endswith('test.conllu')), None)

UD = 'UniversalDependencies'
UD_URL = 'https://api.github.com/repos/UniversalDependencies/{}/contents/'
UD_DNAME = '_UD'
def download_ud(corpus_name, root_dir=None, overwrite=True):
    """Download a corpus ``corpus_name`` from Universal Dependencies.

    :param corpus_name: name of the corpus as it specified on the Universal
                        Dependencies' site
    :type corpus_name: str
    :param root_dir: path to the root storage. If None, default value will
                     be used
    :type root_dir: str
    :param overwrite: False means do not download the corpus if it's already
                      kept in the corpus storage
    """
    fpaths = []
    dpath = os.path.join(root_dir if root_dir else get_root_dir(),
                         CORPUS_DNAME, UD_DNAME)
    fcont = corpus_name + '.contents'
    try:
        download_file(UD_URL.format(corpus_name),
                      dpath=dpath, fname=fcont, overwrite=overwrite,
                      log_msg='Downloading {}/{} contents'
                                  .format(UD, corpus_name))
    except HTTPError as e:
        if e.code == 404:
            print('Corpus {} has not found in Universal Dependencies'
                      .format(corpus_name))
            fpaths = None
        else:
            raise e
    if fpaths is not None:
        train, dev, test = [None] * 3
        fcont = os.path.join(dpath, fcont)
        with open(fcont, 'rt', encoding='utf-8') as f:
            cont = json.loads(f.read())
            urls = [o.get('download_url') for o in cont]
            train, dev, test = _get_ud_train_name(urls), \
                                 _get_ud_dev_name(urls), \
                                _get_ud_test_name(urls)
            dpath = os.path.join(dpath, corpus_name)
            if train:
                fpaths.append(
                    download_corpus('{}/{}:train'.format(UD, corpus_name),
                                    train, dname=dpath, root_dir='',
                                    overwrite=overwrite))
            if dev:
                fpaths.append(
                    download_corpus('{}/{}:dev'.format(UD, corpus_name),
                                    dev, dname=dpath, root_dir='',
                                    overwrite=overwrite))
            if test:
                fpaths.append(
                    download_corpus('{}/{}:test'.format(UD, corpus_name),
                                    test, dname=dpath, root_dir='',
                                    overwrite=overwrite))
        os.remove(fcont)
    return fpaths

def remove_ud(corpus_name, root_dir=None):
    """Remove a corpus ``corpus_name`` of Universal Dependencies.

    :param corpus_name: name of the corpus as it specified on the Universal
                        Dependencies' site
    :type corpus_name: str
    :param root_dir: path to the root storage. If None, default value will
                     be used
    :type root_dir: str
    """
    remove_corpus(os.path.join(UD_DNAME, corpus_name), root_dir=root_dir)

def get_ud_train_path(corpus_name, root_dir=None):
    """Return a name of the train corpus of a corpus ``corpus_name`` of 
    Universal Dependencies"""
    dpath = os.path.join(root_dir if root_dir else get_root_dir(),
                         CORPUS_DNAME, UD_DNAME, corpus_name)
    fname = _get_ud_train_name(os.listdir(dpath))
    return os.path.join(dpath, fname) if fname else None

def get_ud_dev_path(corpus_name, root_dir=None):
    """Return a name of the dev corpus of a corpus ``corpus_name`` of 
    Universal Dependencies"""
    dpath = os.path.join(root_dir if root_dir else get_root_dir(),
                         CORPUS_DNAME, UD_DNAME, corpus_name)
    fname = _get_ud_dev_name(os.listdir(dpath))
    return os.path.join(dpath, fname) if fname else None

def get_ud_test_path(corpus_name, root_dir=None):
    """Return a name of the test corpus of a corpus ``corpus_name`` of 
    Universal Dependencies"""
    dpath = os.path.join(root_dir if root_dir else get_root_dir(),
                         CORPUS_DNAME, UD_DNAME, corpus_name)
    fname = _get_ud_test_name(os.listdir(dpath))
    return os.path.join(dpath, fname) if fname else None

GICR = 'GICR'
GICR_URL = 'https://github.com/dialogue-evaluation/morphoRuEval-2017/raw/master/GIKRYA_texts_new.zip'
GICR_DNAME = 'gicr'
def download_gicr(root_dir=None, overwrite=True):
    """Download the corpus GICR.
    
    :param root_dir: path to the root storage. If None, default value will
                     be used
    :type root_dir: str
    :param overwrite: False means do not download the corpus if it's already
                      kept in the corpus storage
    """
    return download_corpus(GICR, GICR_URL, dname=GICR_DNAME, root_dir=root_dir,
                           overwrite=overwrite, file_noless=8000000)

def remove_gicr(root_dir=None):
    """Remove the corpus GICR.

    :param root_dir: path to the root storage. If None, default value will
                     be used
    :type root_dir: str
    """
    remove_corpus(GICR_DNAME, root_dir=root_dir)

OPENCORPORA = 'OpenCorpora'
OPENCORPORA_URL = 'http://opencorpora.org/files/export/annot/annot.opcorpora.xml.bz2'
OPENCORPORA_NOAMB_URL = 'http://opencorpora.org/files/export/annot/annot.opcorpora.no_ambig.xml.bz2'
OPENCORPORA_NOAMB_NOUNKN_URL = 'http://opencorpora.org/files/export/annot/annot.opcorpora.no_ambig_strict.xml.bz2'
OPENCORPORA_DICT_URL = 'http://opencorpora.org/files/export/dict/dict.opcorpora.xml.bz2'
OPENCORPORA_DNAME = 'opencorpora'
def download_opencorpora(root_dir=None, overwrite=True,
                         noamb=False, nounkn=False):
    """Download the corpus OpenCorpora.
    
    :param root_dir: path to the root storage. If None, default value will
                     be used
    :type root_dir: str
    :param overwrite: False means do not download the corpus if it's already
                      kept in the corpus storage
    :param noamb: download a verstion of the corpus with removed ambiguity
    :param nounkn: download a verstion of the corpus without UNKN tags
                   (used only if *noamb* is True)
    """
    if not noamb:
        nounkn = False
    url = OPENCORPORA_NOAMB_NOUNKN_URL if nounkn else \
          OPENCORPORA_NOAMB_URL if noamb else \
          OPENCORPORA_URL
    return (
        download_corpus(OPENCORPORA + ' 1 of 2', url,
                        dname=OPENCORPORA_DNAME, root_dir=root_dir,
                        overwrite=overwrite, file_noless=1300000 if noamb else
                                                         32000000),
        download_corpus(OPENCORPORA + ' 2 of 2', OPENCORPORA_DICT_URL,
                        dname=OPENCORPORA_DNAME, root_dir=root_dir,
                        overwrite=overwrite, file_noless=16000000)
    )

def remove_opencorpora(root_dir=None):
    """Remove the corpus OpenCorpora.

    :param root_dir: path to the root storage. If None, default value will
                     be used
    :type root_dir: str
    """
    remove_corpus(OPENCORPORA_DNAME, root_dir=root_dir)

RNC = 'RNC'
RNC_URL = 'https://github.com/dialogue-evaluation/morphoRuEval-2017/raw/master/RNC_texts.rar'
RNC_DNAME = 'rnc'
def download_rnc(root_dir=None, overwrite=True):
    """Download the corpus RNC.
    
    :param root_dir: path to the root storage. If None, default value will
                     be used
    :type root_dir: str
    :param overwrite: False means do not download the corpus if it's already
                      kept in the corpus storage
    """
    return download_corpus(RNC, RNC_URL, dname=RNC_DNAME, root_dir=root_dir,
                           overwrite=overwrite, file_noless=5000000)

def remove_rnc(root_dir=None):
    """Remove the corpus RNC.

    :param root_dir: path to the root storage. If None, default value will
                     be used
    :type root_dir: str
    """
    remove_corpus(RNC_DNAME, root_dir=root_dir)

SYNTAGRUS = 'SynTagRus'
SYNTAGRUS_TRAIN_URL = 'https://github.com/UniversalDependencies/UD_Russian-SynTagRus/raw/master/ru_syntagrus-ud-train.conllu'
SYNTAGRUS_DEV_URL = 'https://github.com/UniversalDependencies/UD_Russian-SynTagRus/raw/master/ru_syntagrus-ud-dev.conllu'
SYNTAGRUS_TEST_URL = 'https://github.com/UniversalDependencies/UD_Russian-SynTagRus/raw/master/ru_syntagrus-ud-test.conllu'
SYNTAGRUS_DNAME = '_UD/UD_Russian-SynTagRus'
def download_syntagrus(root_dir=None, overwrite=True):
    """Download the corpus SynTagRus.
    
    :param root_dir: path to the root storage. If None, default value will
                     be used
    :type root_dir: str
    :param overwrite: False means do not download the corpus if it's already
                      kept in the corpus storage
    """
    fpaths = []
    fpaths.append(download_corpus(SYNTAGRUS + ' 1 of 3', SYNTAGRUS_TRAIN_URL,
                                  dname=SYNTAGRUS_DNAME, root_dir=root_dir,
                                  overwrite=overwrite, file_noless=80000000))
    fpaths.append(download_corpus(SYNTAGRUS + ' 2 of 3', SYNTAGRUS_DEV_URL,
                                  dname=SYNTAGRUS_DNAME, root_dir=root_dir,
                                  overwrite=overwrite, file_noless=10000000))
    fpaths.append(download_corpus(SYNTAGRUS + ' 3 of 3', SYNTAGRUS_TEST_URL,
                                  dname=SYNTAGRUS_DNAME, root_dir=root_dir,
                                  overwrite=overwrite, file_noless=10000000))
    return fpaths

def remove_syntagrus(root_dir=None):
    """Remove the corpus SynTagRus.

    :param root_dir: path to the root storage. If None, default value will
                     be used
    :type root_dir: str
    """
    remove_corpus(SYNTAGRUS_DNAME, root_dir=root_dir)


class _AbstractCorpus:
    """Wrapper for a known corpus, to simplify its usage"""

    name = 'AbstractCorpus'
    _dl_params = ''

    @classmethod
    def train(cls):
        raise ValueError('ERROR: {} does not have a train part'
                             .format(cls.name))

    @classmethod
    def dev(cls):
        raise ValueError('ERROR: {} does not have a dev part'
                             .format(cls.name))

    @classmethod
    def test(cls):
        raise ValueError('ERROR: {} does not have a test part'
                             .format(cls.name))

    @classmethod
    def isfile(cls, fpath):
        if not os.path.isfile(fpath):
            raise FileNotFoundError(
                '{} is not found. Download it first with {}({})'
                    .format(cls.name, cls._dl_name, cls._dl_params)
            )


class AdjustedForSpeech(_AbstractCorpus):
    """Wrapper for a known corpus, adjusted for speech"""

    def __init__(self, corpus):
        """
        :param corpus: one of the children of _AbstractCorpus
        """
        #assert issubclass(corpus, _AbstractCorpus), 'ERROR: Unknown input corpus'
        self.name = corpus.name
        if hasattr(corpus, 'train') and callable(corpus.train):
            self.train = lambda: Conllu.fix(corpus.train(),
                                            adjust_for_speech=True)
        if hasattr(corpus, 'dev') and callable(corpus.dev):
            self.dev = lambda: Conllu.fix(corpus.dev(),
                                          adjust_for_speech=True)
        if hasattr(corpus, 'test') and callable(corpus.test):
            self.test  = lambda: Conllu.fix(corpus.test(),
                                            adjust_for_speech=True)


class gicr(_AbstractCorpus):
    """Wrapper for GICR corpus"""

    name = GICR
    _dl_name = 'download_gicr'

    @staticmethod
    def _fix(line):
        line = line.strip()
        if line:
            line = line.split('\t')
            if len(line) == 5:
                line = '{}\t{}\t{}\t{}\t_\t{}\t_\t_\t_\t_' \
                           .format(*[x for x in line])
            else:
                raise ValueError('Invalid format')
        return line

    @classmethod
    def train(cls):
        """Return train part of GICR corpus in CONLL-U format"""
        fpath = get_corpus_fpath(dname=GICR_DNAME, url=GICR_URL)
        cls.isfile(fpath)
        return Conllu.load(read_zip(fpath, 'gikrya_new_train.out',
                                    process_line=cls._fix),
                           log_file=LOG_FILE)

    @classmethod
    def test(cls):
        """Return test part of GICR corpus in CONLL-U format"""
        fpath = get_corpus_fpath(dname=GICR_DNAME, url=GICR_URL)
        cls.isfile(fpath)
        return Conllu.load(read_zip(fpath, 'gikrya_new_test.out',
                                    process_line=cls._fix),
                           log_file=LOG_FILE)


re_gdesc     = re.compile('<grammeme parent="(.*)">'
                         '<name>(.+)</name>.+?'
                         '</grammeme>')
re_text      = re.compile('<text id="([^"]+)" parent="([^"]+)" name="([^"]+)">')
re_tag       = re.compile('<tag>(.+)</tag>')
re_paragraph = re.compile('<paragrapg id="(.+)">')
re_sentence  = re.compile('<sentence id="(.+)">')
re_source    = re.compile('<source>(.+)</source>')
re_token     = re.compile('<token id="[^"]+" text="[^"]+">'
                          '<tfr rev_id="[^"]+" t="([^"]+)">(<v>.+</v>)</tfr>'
                          '</token>')
re_v         = re.compile('<v>(<l .+?></l>)</v>')
re_lemma     = re.compile('<l id="[^"]+" t="([^"]+)">(<g .+?/>)?</l>')
re_grammeme  = re.compile('<g v="([^"]+)"/>')
class opencorpora(_AbstractCorpus):
    """Wrapper for OpenCorpora corpus"""

    name = OPENCORPORA
    _dl_name = 'download_opencorpora'

    @classmethod
    def _fix(cls, line):
        line = line.strip()
        rex = re_token.search(line)
        if rex:
           wform, vv = rex.groups()
           res = []
           for v in re_v.findall(vv):
               lemma, grammemes = re_lemma.search(v).groups()
               feats = OrderedDict()
               pos = '_'
               for g in re_grammeme.findall(grammemes):
                   if g == 'gen1':
                       g = 'gent'
                   if g == 'loc1':
                       g = 'loct'
                   feat = cls._gparents.get(g)
                   if feat == 'POST' or feat is None:
                       pos = g
                   else:
                       feats[feat] = g # may override
               res.append('1\t{}\t{}\t{}\t_\t{}\t_\t_\t_\t_\t'
                              .format(wform, lemma, pos,
                                      '|'.join('='.join((x, feats[x]))
                                                   for x in feats)
                                                       if feats else
                                      '_'))
               break  # in the case of ambuguation we add only 1st version
        else:
            rex = re_text.search(line)
            if rex:
                id_, parent, name = rex.groups()
                res = ['# newdoc id = ' + id_,
                       '# parent_doc_id = ' + parent,
                       '# doc_name = ' + unescape(name)]
            else:
                rex = re_tag.search(line)
                if rex:
                    tag, = rex.groups()
                    res = '# tag_{} = {}'.format(cls._tag_id, unescape(tag))
                    cls._tag_id += 1
                else:
                    rex = re_paragraph.search(line)
                    if rex:
                        id_, = rex.groups()
                        res = '# newpar id = ' + id_
                    else:
                        rex = re_sentence.search(line)
                        if rex:
                            id_, = rex.groups()
                            res = '# sent_id = ' + id_
                        else:
                            rex = re_source.search(line)
                            if rex:
                                source, = rex.groups()
                                res = '# text = ' + unescape(source)
                            elif line in [
                                '</sentence>', '</paragraph>', '</text>'
                            ]:
                                res = ''
                                cls._tag_id = 1
                            else:
                                res = None
        return res

    @classmethod
    def train(cls, noamb=False, nounkn=False):
        """Return OpenCorpora corpus in CONLL-U format.

        :param noamb: download a version of the corpus with removed ambiguity
        :param nounkn: download a version of the corpus without UNKN tags
                       (used only if *noamb* is True)
        """
        if not noamb:
            nounkn = False
        url = OPENCORPORA_NOAMB_NOUNKN_URL if nounkn else \
              OPENCORPORA_NOAMB_URL if noamb else \
              OPENCORPORA_URL
        cls._dl_params =   'noamb=True'  if noamb  else '' \
                       + ', nounkn=True' if nounkn else ''
        gparents = {}
        fpath = get_corpus_fpath(dname=OPENCORPORA_DNAME,
                                 url=OPENCORPORA_DICT_URL)
        for line in read_bz2(fpath):
            line = line.strip()
            if line == '</grammemes>':
                break
            else:
                res = re_gdesc.search(line)
                if res:
                    parent, name = res.groups()
                    gparents[name] = parent if parent else name
        for name, p in gparents.items():
            while True:
                parent = p
                p = gparents.get(p)
                if p == parent:
                    break
            gparents[name] = parent
        cls._gparents = gparents

        fpath = get_corpus_fpath(dname=OPENCORPORA_DNAME, url=url)
        cls.isfile(fpath)
        cls._tag_id = 1
        return Conllu.load(
            (x for x in read_bz2(fpath, process_line=cls._fix)
                 if x is not None),
            log_file=LOG_FILE
        )


class rnc(_AbstractCorpus):
    """Wrapper for RNC corpus"""

    name = RNC
    _dl_name = 'download_rnc'

    @staticmethod
    def _fix(line):
        line = line.strip()
        if line:
            line = line.split('\t')
            if len(line) == 5:
                if line[4] != '_':
                    if line[3] == '_':
                        line[3] = line[4]
                    else:
                        line[3] += '|' + line[4]
                line = '1\t{}\t{}\t{}\t_\t{}\t_\t_\t_\t_' \
                           .format(line[0], line[1], line[2], line[3])
            elif (len(line) == 1 and line[0][0] == '=') \
              or (len(line) == 3 and line[0] == 'PUNCT'):
                line = None
            else:
                raise ValueError('Invalid format', line)
        return line

    @classmethod
    def train(cls):
        """Return RNC corpus in CONLL-U format"""
        fpath = get_corpus_fpath(dname=RNC_DNAME, url=RNC_URL)
        cls.isfile(fpath)
        return Conllu.load(
            (x for x in read_rar(fpath,  'RNCgoldInUD_Morpho.conll',
                                 process_line=cls._fix) if x is not None),
            log_file=LOG_FILE
        )


class syntagrus(_AbstractCorpus):
    """Wrapper for SynTagRus corpus"""

    name = SYNTAGRUS
    _dl_name = 'download_syntagrus'

    @classmethod
    def train(cls):
        fpath = get_corpus_fpath(dname=SYNTAGRUS_DNAME,
                                 url=SYNTAGRUS_TRAIN_URL)
        cls.isfile(fpath)
        return Conllu.load(fpath, log_file=LOG_FILE)

    @classmethod
    def dev(cls):
        fpath = get_corpus_fpath(dname=SYNTAGRUS_DNAME,
                                 url=SYNTAGRUS_DEV_URL)
        cls.isfile(fpath)
        return Conllu.load(fpath, log_file=LOG_FILE)

    @classmethod
    def test(cls):
        fpath = get_corpus_fpath(dname=SYNTAGRUS_DNAME,
                                 url=SYNTAGRUS_TEST_URL)
        cls.isfile(fpath)
        return Conllu.load(fpath, log_file=LOG_FILE)


class UniversalDependencies(_AbstractCorpus):
    """Wrapper for SynTagRus corpus"""

    def __init__(self, corpus_name, root_dir=None):
        self.name = '{}/{}'.format(UD, corpus_name)
        self._dl_name = 'download_ud({})'.format(corpus_name)
        self._root_dir = root_dir
        self._corpus_name = corpus_name

    def isfile(self, fpath):
        if not os.path.isfile(fpath):
            raise FileNotFoundError(
                '{} is not found. Download it first with {}()'
                    .format(cls.name, cls._dl_name)
            )

    def train(self):
        fpath = get_ud_train_path(self._corpus_name, root_dir=self._root_dir)
        self.isfile(fpath)
        return Conllu.load(fpath, log_file=LOG_FILE)

    def dev(self):
        fpath = get_ud_dev_path(self._corpus_name, root_dir=self._root_dir)
        self.isfile(fpath)
        return Conllu.load(fpath, log_file=LOG_FILE)

    def test(self):
        fpath = get_ud_test_path(self._corpus_name, root_dir=self._root_dir)
        self.isfile(fpath)
        return Conllu.load(fpath, log_file=LOG_FILE)



# -*- coding: utf-8 -*-
# Corpuscula project: CONLL-U support
#
# Copyright (C) 2019-present by Sergei Ternovykh
# License: BSD, see LICENSE for details
"""
Full CONLL-U and CONLL-U Plus formats support.
"""
from collections import OrderedDict
from difflib import SequenceMatcher
from io import BufferedReader
import os
import pickle
import sys

from corpuscula.utils import LOG_FILE, print_progress


def id_to_numeric(id_):
    # there will be ValueError if id is incorrect
    return None if id_ is None else \
           int(id_) if id_.isdecimal() else \
           float(id_)


class Conllu:
    """Full CONLL-U and CONLL-U Plus formats support"""

    STD_COLUMNS = ['ID', 'FORM', 'LEMMA', 'UPOS', 'XPOS',
                   'FEATS', 'HEAD', 'DEPREL', 'DEPS', 'MISC']

    def __new__(cls):
        raise NotImplementedError('This class has only static methods')

    @staticmethod
    def fix(corpus, split_multi=False, adjust_for_speech=False):
        """Fix Parsed CONLL-U structure of *corpus* (renew token ids, generate
        sentence ids and text if they are not present in metadata, etc.)

        :param split_multi: if True then wforms with spaces will be processed
                            as multiword tokens
        :param adjust_for_speech: if True, remove all non alphanumeric tokens
                                  and convert all words to lower case
        :rtype: iter
        """
        for sent_no, sentence in enumerate(corpus):
            sentence, sentence_meta = \
                sentence if isinstance(sentence, tuple) else \
                (sentence, OrderedDict())

            tokens = []
            id_, sub_id = 0, 0
            multi_token, last_old_id = None, None
            for token in sentence:
                if adjust_for_speech:
                    wform = token['FORM']
                    if wform and any(x.isalnum() for x in wform):
                        token['FORM'] = wform.lower()
                        if token['LEMMA']:
                            token['LEMMA'] = token['LEMMA'].lower()
                    elif wform is not None:
                        continue
                tokens.append(token)
                columns = token.keys()
                if 'ID' in token and 'FORM' in columns:
                    old_id = token['ID'] or ''
                    if '.' in old_id:
                        sub_id += 1
                        token['ID'] = str(id_) + '.' + str(sub_id)
                    elif multi_token:
                        #if '.' in old_id:
                        #    sub_id += 1
                        #    token['ID'] = str(id_) + '.' + str(sub_id)
                        #else:
                        id_ += 1
                        sub_id = 0
                        token['ID'] = str(id_)
                        if old_id == last_old_id:
                            multi_token['ID'] += '-' + str(id_)
                            multi_token = None
                    else:
                        wform = token['FORM']
                        if wform:
                            id_ += 1
                            sub_id = 0
                            token['ID'] = str(id_)
                            dash_pos = old_id.rfind('-')
                            if dash_pos > 0:
                                multi_token = token
                                last_old_id = old_id[dash_pos + 1:]
                                id_ -= 1
                            else:
                                wforms = wform.strip().split()
                                if split_multi and len(wforms) > 1:
                                    assert not any(
                                        token[x] for x in token.keys()
                                                     if x not in [
                                                         'ID', 'FORM', 'MISC'
                                                     ]
                                    ), 'ERROR: Fix of CONLL-U structure ' \
                                       'with split_multi=True can be made ' \
                                       'before any tagging'
                                    for id_, wform in enumerate(wforms,
                                                                start=id_):
                                        token_ = {}
                                        for column in columns:
                                            token_[column] = \
                                                str(id_) \
                                                    if column == 'ID' else \
                                                wform \
                                                    if column == 'FORM' else \
                                                OrderedDict() if column in [
                                                    'FEATS', 'MISC'
                                                ] else \
                                                None
                                        tokens.append(token_)
                                    token['ID'] += '-' + str(id_)
                        else:
                            sub_id += 1
                            token['ID'] = str(id_) + '.' + str(sub_id)
            sentence = tokens
            if not sentence:
                vals = {}
                for column in columns:
                    vals[column] = \
                        '0.1' if column == 'ID' else \
                        OrderedDict() if column in ['FEATS', 'MISC'] else \
                        None
                sentence = [vals]

            if 'sent_id' not in sentence_meta:
                sentence_meta['sent_id'] = str(sent_no + 1)
                sentence_meta.move_to_end('sent_id', last=False)
            if 'text' not in sentence_meta:
                text = ''
                space_before = False
                ignore_upto = None
                for token in sentence:
                    id_, wform, misc = \
                        token['ID'], token['FORM'], token.get('MISC', {})
                    if ignore_upto:
                        if id_ == ignore_upto:
                            ignore_upto = None
                    elif '.' not in id_:
                        ids = id_.split('-')
                        if len(ids) == 2:
                            id_ = ignore_upto = ids[1]
                        if id_.isdigit():
                            if space_before:
                                text += ' '
                            text += wform
                    space_before = misc.get('SpaceAfter') != 'No'
                sentence_meta['text'] = text

            yield sentence, sentence_meta

    @classmethod
    def from_sentence(cls, wforms, columns=None):
        """Convert a list of wforms to Parsed CONLL-U format.

        :param columns: list of column names. If None, standard CONLL-U columns
                        are used
        :type columns: list(str)

        NB: If wform contains '\u00AD' character, it will be splitted by that
            character, and all it's parts except the last one will be saved
            with the "SpaceAfter=No" MISC attribute"""
        if columns is None:
            columns = cls.STD_COLUMNS
        sent = []
        for i, w in enumerate(wforms, start=1):
            wform = w
            while wform:
                pos = wform.find('\u00AD')
                if pos < 0:
                    w, wform = wform, ''
                else:
                    w, wform = wform[:pos], wform[pos + 1:]
                token = {}
                for column in columns:
                    token[column] = str(i) if column == 'ID' else \
                                    w if column == 'FORM' else \
                                    OrderedDict() if column in [
                                        'FEATS', 'MISC'
                                    ] else \
                                    None
                if wform and 'MISC' in columns:
                    token['MISC']['SpaceAfter'] = 'No'
                sent.append(token)
        return sent

    @classmethod
    def from_sentences(cls, sentences, split_multi=False,
                       adjust_for_speech=False, columns=None):
        """Convert a sequence of tokenized sentences to Parsed CONLL-U format

        :param fix: fix CONLL-U structure of after conversion
        :param split_multi: if True then wforms with spaces will be processed
                            as multiword tokens (used only with fix=True)
        :param adjust_for_speech: if yes, remove all non alphanumeric tokens
                                  and convert all words to lower case (used
                                  only with fix=True)
        :rtype: iter
        """
        yield from cls.fix((cls.from_sentence(s, columns=columns)
                               for s in sentences),
                           split_multi=split_multi,
                           adjust_for_speech=adjust_for_speech)

    @classmethod
    def load(cls, corpus, encoding='utf-8-sig', fix=True, split_multi=False,
             adjust_for_speech=False, log_file=LOG_FILE):
        """Load corpus in CONLL-U format as sequence of Parsed CONLL-U
        sentences. Each sentence returns as tuple of a list of tagged tokens
        and a dict of metadata that can be used to restore corpus back to
        CONLL-U format.

        :param fix: fix CONLL-U structure of after conversion
        :param split_multi: if True then wforms with spaces will be processed
                            as multiword tokens (used only with fix=True)
        :param adjust_for_speech: if yes, remove all non alphanumeric tokens
                                  and convert all words to lower case (used
                                  only with fix=True)
        :return: sentences in Parsed CONLL-U format
        :rtype: sequence of tuple(list(dict(str: str|OrderedDict(str: str))),
                                  OrderedDict(str: str))

        NOTE: For CONLL-U Plus format the field list must be specified in the
              first line of the *corpus* (in the meta variable
              "global.columns")"""
        if fix:
            yield from cls.fix(cls.load(corpus, encoding=encoding, fix=False,
                                        log_file=log_file),
                               split_multi=split_multi,
                               adjust_for_speech=adjust_for_speech)

        else:
            if isinstance(corpus, str):# and os.path.exists(corpus):
                corpus = open(corpus, mode='rt', encoding=encoding)

            if log_file:
                print('Load corpus', file=log_file)
            nsentence = ntoken = 0
            sentence = []
            sentence_meta = OrderedDict()
            columns = None
            sent_no = -1
            for line in corpus:
                line = line.strip()
                if len(line) == 0:
                    if len(sentence) > 0 or len(sentence_meta) > 0:
                        if len(sentence) == 0:
                            if columns is None:
                                columns = cls.STD_COLUMNS
                            vals = {}
                            for column in columns:
                                vals[column] = \
                                    '0.1' if column == 'ID' else \
                                    OrderedDict() \
                                        if column in ['FEATS', 'MISC'] else \
                                    None
                            sentence = [vals]
                        yield sentence, sentence_meta
                        sent_no += 1
                        if log_file and not sent_no % 100:
                            print_progress(sent_no, end_value=None, step=1000,
                                           file=log_file)
                        nsentence += 1
                        ntoken += len(sentence)
                        sentence = []
                        sentence_meta = OrderedDict()
                elif line[0] == '#':
                    meta = tuple(x.strip() for x in line[1:].split('=', 1))
                    sentence_meta.update(
                        [meta if len(meta) > 1 else (meta[0], None)]
                    )
                    if columns is None and meta[0] == 'global.columns':
                        columns = meta[1].split()
                    else:
                        columns = cls.STD_COLUMNS
                else:
                    if columns is None:
                        columns = cls.STD_COLUMNS
                    columns_i = iter(columns)
                    vals = {}
                    for val in line.split('\t'):
                        column = next(columns_i)
                        if column in ['FEATS', 'MISC']:
                            try:
                                val = OrderedDict(
                                    () if val == '_' else
                                    [(k, v) for k, v in [
                                        t.split('=', 1) for t in val.split('|')
                                    ]])
                            except:
                                print('ERROR when loading Conllu', sys.stderr)
                                print(line, sys.stderr)
                                print(column, sys.stderr)
                                print(val, sys.stderr)
                        else:
                            if val == '_':
                                val = None
                            if val == '*' and column not in cls.STD_COLUMNS:
                                val = ''
                        vals[column] = val
                    sentence.append(vals)
            if sentence or sentence_meta:
                yield sentence, sentence_meta
                sent_no += 1
            if log_file and sent_no >= 0:
                print_progress(sent_no + 1, end_value=0, step=1000,
                               file=log_file)
                print('Corpus has been loaded: {} sentences, {} tokens'
                          .format(nsentence, ntoken),
                      file=log_file)

            if isinstance(corpus, BufferedReader):
                corpus.close()

    @classmethod
    def get_as_text(cls, corpus, fix=True, split_multi=False,
                    adjust_for_speech=False, log_file=LOG_FILE):
        """Convert a *corpus* in Parsed CONLL-U format to text form.

        :rtype: iter(str)
        """
        if log_file:
            print('Save corpus', file=log_file)
        sent_no = -1
        for sent_no, sentence in enumerate(
            cls.fix(corpus, split_multi=split_multi,
                    adjust_for_speech=adjust_for_speech) if fix else
            corpus
        ):
            if log_file and not sent_no % 100:
                print_progress(sent_no, end_value=None, step=1000,
                               file=log_file)
            sentence, sentence_meta = \
                sentence if isinstance(sentence, tuple) else \
                (sentence, OrderedDict())

            columns, feats_idx, misc_idx = [None] * 3
            first_meta = []
            for meta, val in sentence_meta.items():
                if columns is None and meta == 'global.columns':
                    columns = val.split()
                    try:
                        feats_idx = columns.index('FEATS')
                    except ValueError:
                        pass
                    try:
                        misc_idx = columns.index('MISC')
                    except ValueError:
                        pass
                line = '# {}{}'.format(meta, ' = {}'.format(val)
                                                 if val is not None else
                                             '')
                if columns is None:
                    first_meta.append(line)
                else:
                    yield line + '\n'
            line = []
            for token in sentence:
                if columns is None:
                    columns = []
                    cols = set(token.keys())
                    for col in cls.STD_COLUMNS:
                        if col in cols:
                            if col == 'FEATS':
                                feats_idx = len(columns)
                            if col == 'MISC':
                                misc_idx = len(columns)
                            columns.append(col)
                            cols.remove(col)
                    if cols:
                        columns.extend(sorted(cols))
                        yield '# global.columns = ' + ' '.join(columns) + '\n'
                    if first_meta:
                        for line in first_meta:
                            yield line + '\n'
                line = ['_' if token[x] is None else
                        '*' if token[x] == ''
                           and x not in cls.STD_COLUMNS else
                        token[x] for x in columns]
                if feats_idx is not None:
                    feats = line[feats_idx]
                    line[feats_idx] = \
                        '|'.join('='.join((x, feats[x])) \
                                     for x in sorted(feats)) \
                                         if feats else \
                                 '_'
                if misc_idx is not None:
                    misc = line[misc_idx]
                    line[misc_idx] = \
                        '|'.join('='.join((x, misc[x])) \
                                     for x in sorted(misc)) \
                                         if misc else \
                                 '_'
                yield '\t'.join(line) + '\n'
            yield '\n'
        if log_file and sent_no >= 0:
            print_progress(sent_no + 1, end_value=0, step=1000,
                           file=log_file)
            print('Corpus has been saved', file=log_file)

    @classmethod
    def save(cls, corpus, file_path, **kwargs):
        """Save a *corpus* in Parsed CONLL-U format to CONLL-U file"""
        with open(file_path, mode='wt', encoding='utf-8') as f:
            for line in cls.get_as_text(corpus, **kwargs):
                print(line, end='', file=f)



# -*- coding: utf-8 -*-
# Corpuscula project: Items Dictionary
#
# Copyright (C) 2019-present by Sergei Ternovykh
# License: BSD, see LICENSE for details
"""
Just a simple wrapper to store some structured data. For example (see lower),
to store dictionary of names.
"""
from copy import deepcopy
import pickle
import sys

from corpuscula.utils import LOG_FILE


class Items:

    def __init__(self, restore_from=None, backup_to=None):
        """
        :param restore_from: path to backup file to load from
        :type restore_from: str
        :param backup_to: path to backup file to save to
        :type backup_to: str
        :param log_file: stream for info messages
        :type file
        """
        self._items = {}  # {item_class: {}}

        if restore_from:
            self.restore_from(restore_from)
        if backup_to:
            self.backup_to(backup_to)

    def backup(self):
        """Get current state"""
        return self._items

    def backup_to(self, file_path):
        """Store current state to file"""
        with open(file_path, 'wb') as f:
            pickle.dump(self.backup(), f, 2)

    def restore(self, o):
        """Restore current state from backup object"""
        self._items = o

    def restore_from(self, file_path):
        """Restore current state from file"""
        with open(file_path, 'rb') as f:
            self.restore(pickle.load(f))

    def isempty(self, item_class=None):
        """Check if current state is not contain any information"""
        return not self._items.get(item_class)

    def get(self, item_class=None, item=None, copy=True):
        """Return a value of some *item* of some class. If item is None, return
        all items of the class.

        :param copy: if True, return a copy of the object (default). Elsewise,
                     return the object itself
        """
        assert item_class in self._items, 'ERROR: Item class does not exist'
        res = self._items[item_class]
        if item:
            res = res.get(item.lower() if isinstance(item, str) else item)
        return deepcopy(res) if copy else res

    def item_isknown(self, item, item_class):
        return self._items.get(item_class, {}).get(
            item.lower() if isinstance(item, str) else item
        ) is not None

    def load(self, items, item_class=None, update=False, encoding='utf-8-sig',
             log_file=LOG_FILE):
        """Load a list of *items* of some class.

        :param items: {item: {attr: val}}, [item], set(item)
        :type items: dict|list|set
        :param item_class: some identificator to designate the list given 
                           (e.g.: 'surname')
        :param update: if a list of the given class is already exists, append
                       or update it
        """
        assert self.isempty(item_class) or update, 'ERROR: Item class is ' \
            'already exists. Use update=True to append or update it'
        if not isinstance(items, dict):
            items = dict.fromkeys(items, {})
        old_len = None if self.isempty(item_class) else len(self._items[item_class])
        items_ = {}
        for item, props in items.items():
            props[None] = item
            items_[item.lower() if isinstance(item, str) else item] = props
        self._items.setdefault(item_class, {}).update(items_)
        print('Loaded {} items{} ({})'
                  .format(len(items_),
                          ' to class "{}"'
                              .format(item_class) if item_class else '',
                          'new' if old_len is None else
                          '{} new, {} total'
                              .format(old_len, len(self._items[item_class]))),
              file=log_file)



# -*- coding: utf-8 -*-
# Corpuscula project
#
# Copyright (C) 2019-present by Sergei Ternovykh
# License: BSD, see LICENSE for details
"""
Corpuscula is a part of the RuMor project. It is a toolkit that simplifies
corpus processing. Highlights are:

* full CONLL-U support (includes CONLL-U Plus)
* wrappers for known corpora of Russian language
* parser and wrapper for russian part of Wikipedia
* Corpus Dictionary that can be used for further morphology processing
"""
from corpuscula._version import __version__
from corpuscula.conllu import Conllu
from corpuscula.corpus_dict import CorpusDict
from corpuscula.items import Items
